# Changelog

## v0.3.10
 - Xoshiro RNGs are now serialized in every snapshot
 - export system improvements for custom/external types. In particular, how a type is reconstructed (other than through the keyword-only constructor) can now be controlled through the new `@importconstruct` macro or an additional argument to `@genexport`.
 - more accurate module context for type name stringification and type reconstruction
 - Makie plotting pipeline moved to new package
 - bugfix in type dissection for legacy data import

## v0.3.9
 - behaviour change: BoxGrids now throw an error if no suitable box can be found for a particle (e.g. when it is outside of the domain) instead of simply ignoring the particle. This should never happen for correctly written models.
 - behaviour change: `ParticleObstacleContainer`s and `POLOContainers` now expect the `SO` type parameter to be `<: AbstractObstacle{N}`
 - new export system convenience functions for dealing with custom types other than particles and parameters
 - bugfix: new particles added from the `step!` function are now folded back into the domain before the next force computation. This may affect simulation results in periodic domains.
 - bugfix: the function `isperiodic` now actually returns correct values
 - bugfix: `readdict` for the JLD2 export backend now produces nested dictionaries instead of a single, flat dictionary with slash-separated paths, consistent with the `HDF5` version
 - minor fixes removing noisy depreciation warnings and debug output

## v0.3.8
 - general cleanup for first public release

## v0.3.7
- dropped support for Julia versions < 1.7
- added optional argument to block execution of finalizers in `evolve!`
- added new TimedCallback
- added export support for Xoshiro RNG
- added exportrecursive macro for recursive use of generated serialization functions
- various improvements in the export system

## v0.3.6
- added a convenience function to save simulationas
- fixed link to changelog in README
- `BoxGrid` now gracefully handles system sizes smaller than the `interactionrange`
- improved improved Makie plotting support
- first steps towards an improved callback interface — callbacks can now have an additional `postpropagate` hook

## v0.3.5
- fixed import of parametric types
- added support for inverted obstacles
- improved Makie plotting support

## v0.3.4
- particles are now `reset!` before the main loop starts
- introduced `autogrid!` function (even more convenient than `addgrid!`!)
- various bugs fixed

## v0.3.3
- import for v0.2 files fixed again

## v0.3.2
- every (generic) export function now has a `warn` flag, which can be used to disable all warnings
- `oldestancestor(id, registrar)` no longer considers IDs with `iszero(id) == true` as valid ancestors.
    This is useful because all cells added by `addparticle!` without a specified parent get registered as
    descendants of `UInt64(0)`
- added `addparticles!` and `addgrid!` function to add multiple particles
- `SimpleParticleContainer` and `ParticleObstacleContainer` can now be used without `BoxGrid`s. Simply add `InPartS.NoBoxing` as a third (fourth) type parameter
- more bugs fixed


## v0.3.1
- various bugs fixed in the export system (thanks to Dorian!)
- reintroduced Particle-type Simulation constructor

## v0.3.0
### Breaking changes
3D systems are now supported! This comes with a bunch of API changes to make things less awkward, to wit
- many basic InPartS types (including abstract types for models) now have a type parameter `D`, indicating their dimensions.
- the World struct has been dissolved:
    - Time is now kept directly in the simulation and can be accessed using the `current_time` and `current_step` functions.
    - Domain geometry is now handled by Domain structs, subtypes of AbstractDomain. InPartS comes with two default domain structs, `Domain2D` and `Domain3D` which do pretty much what you expect them to do. The old `rem` and `fold_back` functions still work, but keep in mind that the field in the simulation struct has been renamed to `sim.domain`.
- Some obstacle definitions (`AxisAlignedWall`, `DiskObstacle`) have been extended to work in 3D. In the case of `DiskObstacle`s, the first type parameter now indicates the dimensions of the Obstacle. Note that for `AxisAlignedWall`s, type parameters have not changed since the memory layout is independent of system dimension.
- `BoxGrid` has been extended to 3D (working with 26-neighbourhoods). The first type parameter now indicates the dimensions of the box grid.
- `AbstractParticleContainer` now has a type parameter `N` which indicates the supported system dimension. `SimpleParticleContainer` and `ParticleObstacleContainer` have also gained a new first type parameter `N`, and now extend `AbstractParticleContainer{N}`.
- the function `Base.ndims` can now be used to determine the supported dimension for `AbstractParticle`s, `AbstractParticleContainer`s, `BoxGrid`s and `AbstractDomain`s without resorting to type parameters.

Aside from the 3D-related stuff, there are some other API changes
- `InPartS.ParticleType(::AbstractParams)` has been removed because
    1. it wasn't actually used anywhere
    2. it enforces a 1-to-1-relation between params and particles, which isn't necessary and/or practical
### Non-breaking changes & new features
 - 3D models are now supported (see https://gitlab.gwdg.de/eDLS/InPartS-models/colonygrowth-models for a working model)
 - callbacks are now called before a new `dt` is computed to make interacting with system dynamics safer and prevent explosions
 - `@exportadd` now actually works, so if you relied on it being broken, you shouldn't have
 - `@exportcollect` has been extended to AdditionalExport fields

## v0.2.4
 - added multithreading (!66)
 - changed behaviour of `serialize!(group, object)` to return `group` (1ad689e5)

## v0.2.3
 - minor bugfixes

## v0.2.2
 - Make use of custom group size in `JLD2` backend
 - Fix type-instability in `closestpoints`. MOAR PERFORMANCE
