using .JLD2

# InPartS IO Backend API
dfopen(type::Type{JLD2.JLDFile}, filename::String, mode::String = "r") = JLD2.jldopen(filename, mode)

dfclose(f::JLD2.JLDFile) = close(f)
function dfrename(::Type{JLD2.JLDFile}, oldname, newname; force = false)
    cp(oldname, newname; force = force)
    rm(oldname)
end

gcreate(g::Union{JLD2.JLDFile, JLD2.Group}, name::String; kwargs...) = JLD2.Group(g, name; kwargs...)

isgroup(g::Union{JLD2.JLDFile, JLD2.Group}, key) = (g[key] isa JLD2.Group)
Base.length(g::Union{JLD2.JLDFile, JLD2.Group}) = length(keys(g))

# For backwards compatibility with HDF5 write all SArrays as Arrays
Base.setindex!(g::JLD2.Group, value::SArray, key::String) = setindex!(g, collect(value), key)
Base.setindex!(g::JLD2.JLDFile, value::SArray, key::String) = setindex!(g, collect(value), key)
Base.setindex!(g::JLD2.Group, value::Dict{String, Any}, key::String) = InPartS.writedict(g, value; name = key)
Base.setindex!(g::JLD2.JLDFile, value::Dict{String, Any}, key::String) = InPartS.writedict(g, value; name = key)



# more efficient readdict
# copied and modified from previous JLD2.loadtodict!(Dict{String,Any}(), g)
# to preserve nestedness
# is faster than generic version in generic.jl because reading via `g[k]`
# only happens once
function readdict(g::Union{JLD2.JLDFile, JLD2.Group}; kwargs...)
    d = Dict{String, Any}()
    for k in keys(g)
        v = g[k]
        if v isa JLD2.Group
            d[k] = readdict(v)
        else
            d[k] = v
        end
    end
    return d
end
    
readdict(d::Dict) = d
